/*
 * ABC plugin
 *
 * This plugin converts ABC files to embedded SVG images in WEB browsers.
 *
 * Copyright (C) 2012-2015 Jean-François Moine (http://moinejf.free.fr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */
//#define DEBUG 1

#include "npfunctions.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

//should be in stddef.h..
#ifndef offsetof
#define offsetof(st, m) __builtin_offsetof(st, m)
#endif

#ifdef DEBUG
#define log(fmt, args...) do{printf("%d ", instance->seq); printf(fmt, ##args);} while(0)
#else
#define log(fmt, args...)
#endif

#define PLUGIN_NAME        "ABC plug-in"
#define PLUGIN_DESCRIPTION PLUGIN_NAME " (for abcm2ps >= 7.6.8.)"
#define PLUGIN_VERSION     "1.0.2"

struct Instance {
	NPP npp;
	uint16_t mode;
#ifdef DEBUG
	uint16_t seq;
#endif
	char *abcm2ps;			/* abcm2ps parameters */
};

/* browser functions */
#define NPN_CreateObject(n, c) npnf->createobject(n, c)
#define NPN_RetainObject(o) npnf->retainobject(o)
#define NPN_ReleaseObject(o) npnf->releaseobject(o)
#define NPN_UTF8FromIdentifier(i) npnf->utf8fromidentifier(i)
#define NPN_MemAlloc(sz) npnf->memalloc(sz)
#define NPN_MemFree(p) npnf->memfree(p)
#define NPN_SetException(o, s) npnf->setexception(o, s)
#define NPN_NewStream(m, t, w, s) npnf->newstream(m, t, w, s)
#define NPN_DestroyStream(n, s, r) npnf->destroystream(n, s, r)
#define NPN_Write(n, s, l, b) npnf->write(n, s, l, b)
#define NPN_InvokeDefault(n, o, a, c, r) npnf->invokeDefault(n, o, a, c, r)
#define NPN_Invoke(n, o, i, a, c, r) npnf->invoke(n, o, i, a, c, r)
#define NPN_GetURL(n, u, w) npnf->geturl(n, u, w)
#define NPN_GetURLNotify(n, u, w, d) npnf->geturlnotify(n, u, w, d)
#define NPN_GetValue(n, v, d) npnf->getvalue(n, v, d)
#define NPN_SetValue(n, v, d) npnf->setvalue(n, v, d)
#define NPN_Evaluate(n, o, s, r) npnf->evaluate(n, o, s, r)
#define NPN_ReleaseVariantValue(v) npnf->releasevariantvalue(v)
#define NPN_GetStringIdentifier(s) npnf->getstringidentifier(s)
#define NPN_GetProperty(n, o, p, r) npnf->getproperty(n, o, p, r)

static NPNetscapeFuncs *npnf = NULL;

/* lock for abcmimg definition */
static int lock;

#ifdef DEBUG
static int seq;
#endif
/* (for debug) */
static struct Instance null_inst;
static NPP_t null_npp = {
	.pdata = &null_inst
};

/* create the javascript function 'abcmimg' if it does not exist */
static void abcmimg_create(NPP npp, int mode)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);
#endif
	NPObject *windowObject;
	NPIdentifier abcimg;
	NPVariant result;
	NPString str;
	int no_abcimg;

	/* the test and creation must be protected in case of many threads */
	while (lock != 0)
		usleep(200000);		/* 0.2 s */
	lock = 1;

	NPN_GetValue(npp, NPNVWindowNPObject, &windowObject);
	abcimg = NPN_GetStringIdentifier("abcmimg");
	no_abcimg = 0;
	if (mode == NP_EMBED
	 && NPN_GetProperty(npp, windowObject, abcimg, &result)) {
		switch (result.type) {
		case NPVariantType_Void:
			no_abcimg = 1;
			break;
		case NPVariantType_Object:
			break;
		default:
			log("abcmimg: %d\n", result.type);
			break;
		}
		NPN_ReleaseVariantValue(&result);
	} else {
		no_abcimg = 1;
	}

	/* the function abcmimg replaces the plugin by the images */
	if (no_abcimg) {
		log("building abcmimg\n");
		str.UTF8Characters =
			"function abcmimg(elt, img) {"
				"elt.outerHTML = img;"
			"}";
		str.UTF8Length = strlen(str.UTF8Characters);
		if (NPN_Evaluate(npp, windowObject, &str, &result))
			NPN_ReleaseVariantValue(&result);
		else
			log("evaluate error\n");
	}
	lock = 0;
	NPN_ReleaseObject(windowObject);
}

/* base64 and %xx decoding */
static void b64_dec(FILE *abc_f, char *p)
{
	int i, j, t, v;
	static const signed char b64_dec[80] = {	/* idx = c - 43 */
		62, -1, -1, -1, 63, 52, 53, 54,
		55, 56, 57, 58, 59, 60, 61, -1,
		-1, -1, 64, -1, -1, -1,  0,  1,
		 2,  3,  4,  5,  6,  7,  8,  9,
		10, 11, 12, 13, 14, 15, 16, 17,
		18, 19, 20, 21, 22, 23, 24, 25,
		-1, -1, -1, -1, -1, -1, 26, 27,
		28, 29, 30, 31, 32, 33, 34, 35,
		36, 37, 38, 39, 40, 41, 42, 43,
		44, 45, 46, 47, 48, 49, 50, 51
	};

	i = j = v = 0;
	for ( ; *p != '\0'; p++) {
		unsigned char c;

		c = (unsigned char) *p - 43;
		if (c >= 80)
			continue;
		t = b64_dec[c];
		if (t < 0)
			continue;
		if (t == 64) {		/* '=' */
			j++;
			t = 0;
		}
		v = (v << 6) | t;
		if (++i >= 4) {
			i = 0;
			if (j == 0) {
				fputc(v >> 16, abc_f);
				fputc((v >> 8) & 0xff, abc_f);
				fputc(v & 0xff, abc_f);
			} else {
				if (j <= 2)
					fputc(v >> 16, abc_f);
				if (j <= 1)
					fputc((v >> 8) & 0xff, abc_f);
			}
		}
	}
	if (i != 0) {
		v <<= 8 * (3 - i);
		if (j <= 2)
			fputc(v >> 16, abc_f);
		if (j <= 1)
			fputc((v >> 8) & 0xff, abc_f);
		if (j == 0)
			fputc(v & 0xff, abc_f);
	}
}

static void std_dec(FILE *abc_f, char *p)
{
	unsigned char c, v;

	for ( ; *p != '\0'; p++) {
		c = (unsigned char) *p;
		if (c != '%') {
			fputc(c, abc_f);
			continue;
		}
		c = (unsigned char) *++p;
		if ((unsigned) (c - '0') < 10)
			v = (unsigned) (c - '0');
		else if ((unsigned) (c - 'A') < 6)
			v = (unsigned) (c - 'A' + 10);
		else if ((unsigned) (c - 'a') < 6)
			v = (unsigned) (c - 'a' + 10);
		else
			break;
		v <<= 4;
		c = (unsigned char) *++p;
		if ((unsigned) (c - '0') < 10)
			v += (unsigned) (c - '0');
		else if ((unsigned) (c - 'A') < 6)
			v += (unsigned) (c - 'A' + 10);
		else if ((unsigned) (c - 'a') < 6)
			v += (unsigned) (c - 'a' + 10);
		else
			break;
		fputc(v, abc_f);
	}
}

/* NPP */
NPError NPP_New(NPMIMEType pluginType,
		NPP npp,
		uint16_t mode,
		int16_t argc,
		char *argn[],
		char *argv[],
		NPSavedData *saved)
{
	struct Instance *instance;
	int i;
	char *p, *abc;

	// set up our instance
	instance = (struct Instance *) malloc(sizeof(struct Instance));
	if (!instance)
		return NPERR_OUT_OF_MEMORY_ERROR;
	memset(instance, 0, sizeof *instance);
	instance->npp = npp;
	npp->pdata = instance;
	instance->mode = mode;		/* NP_EMBED or NP_FULL */
#ifdef DEBUG
	instance->seq = ++seq;
	log("New '%s' mode %d argc: %d\n", pluginType, mode, argc);
#endif

	abc = NULL;
	for (i = 0; i < argc; i++) {
		log("   %s = %s\n", argn[i], argv[i]);
		if (strcmp(argn[i], "abc") == 0)
			abc = argv[i];
		if (strcmp(argn[i], "abcm2ps") == 0)
			instance->abcm2ps = strdup(argv[i]);
	}

	/* check if the abcm2ps argument is secure */
	p = instance->abcm2ps;
	if (p) {
		while (*p != '\0') {
			char c;

			if (isalnum((unsigned char) *p))
				;
			else
			     switch (*p) {
			case ' ':
			case '-':
			case '.':
				break;
			case '\t':
			case '\n':
			case '\r':
				*p = ' ';	// this occurs with .html
				break;
			case '\'':
			case '"':
				c = *p;
				goto skip;
			case '<':
				c = '>';
				goto skip;
			case '{':
				c = '}';
skip:
				do {
					p++;
					if (*p == '\\') {
						p++;
						if (*p == c)
							continue;
					}
					if (*p == '\0') {
						log("bad abcm2ps argument\n");
						return NPERR_INVALID_PARAM;
					}
				} while (*p != c);
				break;
			default:
				log("bad abcm2ps char 0x%02x\n", *p);
				return NPERR_INVALID_PARAM;
			}
			p++;
		}
	}

	/* create the javascript function 'abcmimg' if it does not exist */
	abcmimg_create(npp, mode);

	/* if the ABC source is present, generate now */
	if (abc) {
		FILE *abc_f;
		char fn[] = "/tmp/abcmXXXXXX";

		mkstemp(fn);
		abc_f = fopen(fn, "w");
		if (!abc_f)
			return NPERR_OUT_OF_MEMORY_ERROR;
		fwrite(abc, 1, strlen(abc), abc_f);
		fclose(abc_f);
		NPP_StreamAsFile(npp, 0, fn);
		unlink(fn);
	}
	return NPERR_NO_ERROR;
}

NPError NPP_Destroy(NPP npp,
		NPSavedData **save)
{
	struct Instance *instance = (struct Instance *) (npp->pdata);

	log("Destroy\n");
	if (instance->abcm2ps)
		free(instance->abcm2ps);
	free(instance);
	return NPERR_NO_ERROR;
}

NPError NPP_SetWindow(NPP npp,
			NPWindow *window)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);

	log("Setwindow\n");
#endif
	return NPERR_NO_ERROR;
}

NPError NPP_NewStream(NPP npp,
			NPMIMEType type,
			NPStream *stream,
			NPBool seekable,
			uint16_t *stype)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);

	log("NewStream %s %s\n", type, stream->url);
#endif
	/* special case for data: */
	if (strncmp(stream->url, "data:", 5) == 0) {
		FILE *abc_f;
		char *p, *base64;
		char fn[] = "/tmp/abcmXXXXXX";

		p = strchr((char *) stream->url, ',');
		if (!p)
			return NPERR_INVALID_URL;
		p++;
		base64 = strstr((char *) stream->url, "base64");
		mkstemp(fn);
		abc_f = fopen(fn, "w");
		if (!abc_f)
			return NPERR_OUT_OF_MEMORY_ERROR;
		if (base64)
			b64_dec(abc_f, p);
		else
			std_dec(abc_f, p);
		fclose(abc_f);
		NPP_StreamAsFile(npp, 0, fn);
		unlink(fn);
	}

	/* move to a local file */
	*stype = NP_ASFILEONLY;
	return NPERR_NO_ERROR;
}

NPError NPP_DestroyStream(NPP npp,
			NPStream *stream,
			NPReason reason)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);

	log("DestroyStream\n");
#endif
	return NPERR_GENERIC_ERROR;
}


int32_t NPP_WriteReady(NPP npp,
			NPStream* stream)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);

	log("WriteReady\n");
#endif
	return 0;
}

int32_t NPP_Write(NPP npp,
		NPStream *stream,
		int32_t offset,
		int32_t len,
		void *buffer)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);

	log("Write\n");
#endif
	return 0;
}

static int error_append(char **p_image, int image_sz,
			FILE *ferr, size_t err_sz, int tail_sz)
{
	char *p;

	*p_image = realloc(*p_image, image_sz
					+ 14		/* "<pre><![CDATA[" */
					+ err_sz
					+ 9		/* "]]>/<pre>" */
					+ tail_sz);
	p = *p_image;
	if (!p)
		return image_sz;
	p += image_sz;
	strcpy(p, "<pre><![CDATA[");
	p += 14;
	fread(p, 1, err_sz, ferr);
	p += err_sz;

	strcpy(p, "]]></pre>");
	p += 9;
	return p - *p_image;
}

void NPP_StreamAsFile(NPP npp,
			NPStream *stream,
			const char *fname)
{
	struct Instance *instance = (struct Instance *) (npp->pdata);
	NPObject *windowObject, *pluginObject;
	NPIdentifier abcimg;
	NPError err;
	NPVariant result;
	NPVariant args[3];
	char tmp[256], fn_err[] = "/tmp/abcmXXXXXX";
	FILE *fout, *ferr = NULL;
	char *image;
	int image_sz;
	size_t err_sz;
	char out_type;

	out_type = instance->mode == NP_FULL
			? 'X'	/* one image per page */
			: 'g';	/* one image per tune */
	if (instance->abcm2ps) {
		if (strstr(instance->abcm2ps, "-X"))
			out_type = 'X';		/* force XHTML */
		if (strstr(instance->abcm2ps, "-v"))
			out_type = 'v';		/* force SVG one image per page */
	}

	/* convert the ABC tune to svg with abcm2ps */
	mkstemp(fn_err);
	snprintf(tmp, sizeof tmp, "abcm2ps %s %s%s -q%cO- 2> %s",
			fname,
			instance->abcm2ps ? instance->abcm2ps : "",
			instance->mode == NP_FULL ? " --bgcolor white" : "",
			out_type,
			fn_err);
	log("StreamAsFile %s\n", tmp);
	fout = popen(tmp, "r");
	if (!fout) {
		log("popen error");
		goto err1;
	}

	image_sz = 0;
	image = malloc(8192);
	if (!image) {
		log("out of memory\n");
		goto err1;
	}

	/* get the images */
	for (;;) {
		int l;

		l = fread(&image[image_sz], 1, 8192, fout);
		image_sz += l;
		if (l != 8192)
			break;
		image = realloc(image, image_sz + 8192);
		if (!image) {
			log("out of memory\n");
			goto err1;
		}
	}
	pclose(fout);
	fout = NULL;

	/* if some errors add them before or after the images */
	ferr = fopen(fn_err, "r");
	if (ferr) {
		fseek(ferr, 0L, SEEK_END);
		err_sz = ftell(ferr);
		rewind(ferr);
		if (err_sz > 0) {
			switch (out_type) {
			case 'g':		/* tune images */
				image_sz = error_append(&image, image_sz,
							ferr, err_sz,
							0);
				if (!image) {
					log("out of memory\n");
					goto err1;
				}
				break;
			case 'v': {
				char *new_image, *p, *q;

				new_image = malloc(14 + 9 + 1 + err_sz + image_sz);
				if (!new_image) {
					log("out of memory\n");
					goto err1;
				}
				strcpy(new_image, "<pre><![CDATA[");
				p = q = new_image + 14;
				fread(p, 1, err_sz, ferr);
				p += err_sz;
				strcpy(p, "]]></pre>");
				p += 9;
				memcpy(p, image, image_sz + 1);
				image_sz += p - new_image;
				image = new_image;
				break;
			    }
			default:		/* XHTML */
				image_sz -= 7 + 1 + 7 + 1;	/* "<body>\n<html>" */
				image_sz = error_append(&image, image_sz,
							ferr, err_sz,
							4 + 7 + 1 + 7);
				if (!image) {
					log("out of memory\n");
					goto err1;
				}
				strcpy(image + image_sz, "</body>\n</html>");
				image_sz += 7 + 1 + 7;
				break;
			}
		}
		fclose(ferr);
		ferr = NULL;
	}

	/* get the window and plugin objects */
	err = NPN_GetValue(npp, NPNVWindowNPObject, &windowObject);
	if (err != NPERR_NO_ERROR) {
		log("getvalue window obj error %d\n", err);
		goto err2;
	}

	err = NPN_GetValue(npp, NPNVPluginElementNPObject, &pluginObject);
	if (err != NPERR_NO_ERROR) {
		log("getvalue plugin elt obj error %d\n", err);
		goto err3;
	}

	/* invoke the function abcimg */
	abcimg = NPN_GetStringIdentifier("abcmimg");
	OBJECT_TO_NPVARIANT(pluginObject, args[0]);
	STRINGN_TO_NPVARIANT(image, image_sz, args[1]);
	if (NPN_Invoke(npp, windowObject, abcimg, args, 2, &result))
		NPN_ReleaseVariantValue(&result);

	NPN_ReleaseObject(pluginObject);
err3:
	NPN_ReleaseObject(windowObject);
err2:
	free(image);
err1:
	if (fout)
		pclose(fout);
	if (ferr)
		fclose(ferr);
	unlink(fn_err);
}

void NPP_Print(NPP npp,
		NPPrint *platformPrint)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);

	log("Print\n");
#endif
}

int16_t NPP_HandleEvent(NPP npp,
			void *event)
{
#if 0 //def DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);
	XEvent *nativeEvent = (XEvent *) event;

	log("HandleEvent %d\n", nativeEvent->type);
#endif
	return 0;
}

void NPP_URLNotify(NPP npp,
		const char *URL,
		NPReason reason,
		void *notifyData)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);

	log("URLNotify %s\n", URL);
#endif
}

static NPError getValue(NPP npp,
		NPPVariable variable,
		void *value)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);
#endif

	switch (variable) {
	case NPPVpluginNameString:
		log("getValue name\n");
		*((char **) value) = PLUGIN_NAME;
		break;
	case NPPVpluginDescriptionString:
		log("getValue description\n");
		*((char **) value) = PLUGIN_DESCRIPTION;
		break;
#if 1 // defined(XULRUNNER_SDK)
	case NPPVpluginNeedsXEmbed:
		log("getvalue xembed\n");
		*((NPBool *) value) = false;
		break;
#endif
#if 1
	case NPPVpluginScriptableNPObject: {
		NPObject *pluginObject;

		log("getvalue plugin obj\n");
		NPN_GetValue(npp, NPNVPluginElementNPObject, &pluginObject);
		*((NPObject **) value) = 0;
		break;
	   }
#endif
	default:
		log("getValue %d\n", variable);
		return NPERR_INVALID_PARAM;
	}
	return NPERR_NO_ERROR;
}

NPError NPP_GetValue(NPP npp,
		NPPVariable variable,
		void *value)
{
	return getValue(npp, variable, value);
}

NPError NPP_SetValue(NPP npp,
		NPNVariable variable,
		void *value)
{
#ifdef DEBUG
	struct Instance *instance = (struct Instance *) (npp->pdata);

	log("SetValue %d\n", variable);
#endif
	return NPERR_GENERIC_ERROR;
}

/* initialize the NPAPI */
NP_EXPORT(NPError) NP_Initialize(NPNetscapeFuncs *bFuncs,
				NPPluginFuncs *pFuncs)
{
	// Check the size of the provided structure based on the offset of the
	// last member we need.
	if (pFuncs->size < (offsetof(NPPluginFuncs, setvalue) + sizeof(void*)))
		return NPERR_INVALID_FUNCTABLE_ERROR;

#ifdef DEBUG
	printf("0 initialize\n");
#endif
	pFuncs->newp = NPP_New;
	pFuncs->destroy = NPP_Destroy;
	pFuncs->setwindow = NPP_SetWindow;
	pFuncs->newstream = NPP_NewStream;
	pFuncs->destroystream = NPP_DestroyStream;
	pFuncs->asfile = NPP_StreamAsFile;
	pFuncs->writeready = NPP_WriteReady;
	pFuncs->write = NPP_Write;
	pFuncs->print = NPP_Print;
	pFuncs->event = NPP_HandleEvent;
	pFuncs->urlnotify = NPP_URLNotify;
	pFuncs->getvalue = NPP_GetValue;
	pFuncs->setvalue = NPP_SetValue;

	npnf = bFuncs;

	return NPERR_NO_ERROR;
}

NP_EXPORT(char *) NP_GetPluginVersion()
{
	return PLUGIN_VERSION;
}

NP_EXPORT(const char *) NP_GetMIMEDescription()
{
	return "text/vnd.abc:abc:ABC notation";
}

NP_EXPORT(NPError) NP_GetValue(void *future,
				NPPVariable aVariable,
				void *aValue)
{
	return getValue(&null_npp, aVariable, aValue);
}

NP_EXPORT(NPError) NP_Shutdown()
{
#ifdef DEBUG
	struct Instance *instance = &null_inst;

	log("Shutdown\n");
#endif
	return NPERR_NO_ERROR;
}
